local API = GetLibResearchAPIWrapper()
local debug = GetLibResearchDebug()
local RPC = GetLibResearchResearchProgressContainer()
local RPI = GetLibResearchResearchProgressInfo()
local IC = GetLibResearchInventoryContainer()
local MDC = GetLibResearchMotifRefDataContainer()
local MKC = GetLibResearchMotifKnowledgeContainer()
local Config = GetLibResearchConfig()

local CharacterDataManager = { }

function CharacterDataManager:new()
	local obj = {}
	setmetatable(obj, self)
	self.__index = self

	obj.Characters = {}
	obj.CustomData = nil
	obj.CharactersToCompareTo = {}
	obj.ResearchProgressContainers = {}
	obj.MotifKnowledgeContainers = {}
	obj.InventoryContainers = {}
	obj.SelectedCharacter = nil

	return obj;
end

function CharacterDataManager:reSetMetaTable(obj)
	if (obj == nil) then
		return
	end

	setmetatable(obj, self)
	self.__index = self

	for k,v in pairs(obj.Characters) do
		if (obj.ResearchProgressContainers ~= nil) then
			RPC:reSetMetaTable(obj.ResearchProgressContainers[k])
		end
		if (obj.InventoryContainers ~= nil) then
			IC:reSetMetaTable(obj.InventoryContainers[k])
		end
		if (obj.MotifKnowledgeContainers ~= nil) then
			MKC:reSetMetaTable(obj.MotifKnowledgeContainers[k])
		end
	end	
end

function CharacterDataManager:getCharacterNames()
	local result = {}
	for charName,_ in pairs(self.Characters) do
		table.insert(result, charName)
	end
	
	return result;
end

function CharacterDataManager:listOtherCharacters(characterToSkip)
	local result = {}
	if (characterToSkip == nil) then
		characterToSkip = API.GetCharacterName()
	end
	
	debug.debug("listOtherCharacters()")
	for charName,_ in pairs(self.Characters) do
		debug.debug("listOtherCharacters(): " .. charName)
		if (charName ~= characterToSkip) then
			table.insert(result, charName)
		end
	end
	
	return result
end

function CharacterDataManager:refreshCurrentCharacterData(researchRefDataContainer, motifRefDataContainer, bankContainer)
	self:refreshCurrentCharacterResearchData(researchRefDataContainer)
	self:refreshCurrentCharacterMotifData(motifRefDataContainer)
	self:refreshCurrentCharacterInventoryData(bankContainer)
end

function CharacterDataManager:refreshCurrentCharacterMotifData(motifRefDataContainer)
	local charName = API.GetCharacterName()
	self.Characters[charName] = 1

	if (self.MotifKnowledgeContainers == nil) then
		self.MotifKnowledgeContainers = {}
	end
	
	local mkc = self.MotifKnowledgeContainers[charName]
	if (mkc == nil) then
		mkc = MKC:new(charName)
		self.MotifKnowledgeContainers[charName] = mkc
	end
	mkc:refresh(motifRefDataContainer)
end


function CharacterDataManager:refreshCurrentCharacterResearchData(researchRefDataContainer)
	local charName = API.GetCharacterName()
	self.Characters[charName] = 1

	if (self.ResearchProgressContainers == nil) then
		self.ResearchProgressContainers = {}
	end

	local rpc = self.ResearchProgressContainers[charName]
	if (rpc == nil) then
		rpc = RPC:new(charName)
		self.ResearchProgressContainers[charName] = rpc
	end
	rpc:initFromResearchRefDataContainer(researchRefDataContainer)
	rpc:refreshAll()
end

function CharacterDataManager:refreshCurrentCharacterInProgressResearchData()
	local charName = API.GetCharacterName()
	self.Characters[charName] = 1

	if (self.ResearchProgressContainers == nil) then
		self.ResearchProgressContainers = {}
	end

	local rpc = self.ResearchProgressContainers[charName]
	if (rpc == nil) then
		rpc = RPC:new(charName)
		self.ResearchProgressContainers[charName] = rpc
		rpc:refreshAll()
	else
		rpc:refreshInProgress()
	end
end

function CharacterDataManager:refreshCurrentCharacterInventoryData(bankContainer)
	local charName = API.GetCharacterName()
	self.Characters[charName] = 1

	if (self.InventoryContainers == nil) then
		self.InventoryContainers = {}
	end

	local ic = self.InventoryContainers[charName]
	if (ic == nil) then
		ic = IC:new(charName, BAG_BACKPACK)
		self.InventoryContainers[charName] = ic
	end
	
	local result = ic:refresh()
	bankContainer:refresh()

	return result
end

function CharacterDataManager:getComparisonCharacters(addCurrent)
	if (addCurrent == nil) then
		addCurrent = false
	end
	
	local result = {}
	local targetCharacter = API.GetCharacterName()
	if (addCurrent) then
		table.insert(result, targetCharacter)
	end
	
	if ((self.CharactersToCompareTo ~= nil) and (self.CharactersToCompareTo[targetCharacter] ~= nil)) then
		for i = 1,#self.CharactersToCompareTo[targetCharacter] do
			table.insert(result, self.CharactersToCompareTo[targetCharacter][i])
		end
	end
	
	return result
end

function CharacterDataManager:setComparisonCharacters(characters, targetCharacter)
	if (characters == nil) then
		characters = {}
	end
	if (targetCharacter == nil) then
		targetCharacter = API.GetCharacterName()
	end

	debug.debug("setComparisonCharacters: " .. targetCharacter)	
	self.CharactersToCompareTo[targetCharacter] = characters
end

function CharacterDataManager:getResearchProgressInfoOfCurrentCharacter(researchKey)
	local charName = API.GetCharacterName()
	local rpc = self.ResearchProgressContainers[charName]
	return rpc:getByResearchKey(researchKey)
end

function CharacterDataManager:getResearchProgressInfoOfCharacter(character, researchKey)
	local rpc = self.ResearchProgressContainers[character]
	return rpc:getByResearchKey(researchKey)
end

function CharacterDataManager:hasMultipleItemsForTheSameTrait(character, researchKey, existingResults)
	local ic = self.InventoryContainers[character]
	return ic:hasMultipleItemsForTheSameTrait(researchKey, existingResults)
end

function CharacterDataManager:haveMultipleItemsForTheSameTrait(researchKey, existingResults)
	if (existingResults == nil) then
		existingResults = {}
	end
	
	for k,ic in pairs(self.InventoryContainers) do
		existingResults = ic:hasMultipleItemsForTheSameTrait(researchKey, existingResults)
	end
	
	return existingResults
end


function CharacterDataManager:getCustomData()
	return self.CustomData
end

function CharacterDataManager:setCustomData(obj)
	self.CustomData = obj
end

function CharacterDataManager:getSelectedCharacter()
	return self.SelectedCharacter
end

function CharacterDataManager:selectCharacter(characterName)
	self.SelectedCharacter = characterName
end

function CharacterDataManager:getRPCOfSelectedCharacter()
	return self.ResearchProgressContainers[self.SelectedCharacter]
end

function CharacterDataManager:getRPCOfCharacter(characterName)
	return self.ResearchProgressContainers[characterName]
end

function CharacterDataManager:getRPCsOfComparisonCharacter(otherCharacters)
	local chars = self:getSelectedCurrentAndComparisonCharacters(otherCharacters)
	
	local result = {}
	for i = 1,#chars do
		result[chars[i]] = self.ResearchProgressContainers[chars[i]]
	end
	
	return result
end

function CharacterDataManager:refreshRPCOfSelectedAndComparisonCharacters()
	local chars = self:getSelectedCurrentAndComparisonCharacters(true)
	for i = 1, #chars do
		self.ResearchProgressContainers[chars[i]]:refreshAll()
	end
end

function CharacterDataManager:getSelectedCurrentAndComparisonCharacters(otherCharacters)
	local result = {}
	local tmp = {}
	
	table.insert(result, self.SelectedCharacter)
	tmp[self.SelectedCharacter] = 1
	if (otherCharacters) then
		local compChars = self:getComparisonCharacters(true)
		for i = 1, #compChars do
			local char = compChars[i]
			if (tmp[char] == nil) then
				table.insert(result, char)
				tmp[char] = 1
			end
		end
	end
	
	return result
end

function CharacterDataManager:refreshRPCs(otherCharacters, inProgressOnly)
	local rpc = nil
	if (not otherCharacters) then
		rpc = self:getRPCOfSelectedCharacter()
		if (inProgressOnly) then
			rpc:refreshInProgress()
		else
			rpc:refreshAll()
		end
	else
		local chars = self:getSelectedCurrentAndComparisonCharacters(otherCharacters)
		for i = 1,#chars do
			rpc = self:getRPCOfCharacter(chars[i])
			if (inProgressOnly) then
				rpc:refreshInProgress()
			else
				rpc:refreshAll()
			end
		end
	end
end

function CharacterDataManager:refreshInProgressRPCOfSelectedAndComparisonCharacters()
	local chars = self:getComparisonCharacters(true)
	for i = 1, #chars do
		self.ResearchProgressContainers[chars[i]]:refreshInProgress()
	end
end

function CharacterDataManager:getInProgressResearchData(showOtherCharacters)
	local rpcs = RB2.CDM:getRPCsOfComparisonCharacter(showOtherCharacters)
	local researches = {}
	local allMaxResearches = {}
	local currResearches = {}
	local charResNums = {}
	for char, rpc in pairs(rpcs) do
		local tmp = rpc:listInProgressResearches(nil, false)
		local charTable = {}
		for i = 1, #tmp do
			table.insert(researches, tmp[i])
		end
		for i = 1, #Config.TradeSkillTypes do
			local tst = Config.TradeSkillTypes[i].Type
			local actRes = #rpc:listInProgressResearches(tst, false)
			local maxRes = rpc:getMaxResearches(tst)
			charTable[tst] = { maxRes, actRes }
			if (allMaxResearches[tst] == nil) then
				allMaxResearches[tst] = maxRes;
			else
				allMaxResearches[tst] = allMaxResearches[tst] + maxRes;
			end
			if (currResearches[tst] == nil) then
				currResearches[tst] = actRes;
			else
				currResearches[tst] = currResearches[tst] + actRes;
			end
		end		
		charResNums[char] = charTable
	end
	
	return researches, charResNums, allMaxResearches, currResearches
end

function CharacterDataManager:getICOfSelectedCharacter()
	return self.InventoryContainers[self.SelectedCharacter]
end

function CharacterDataManager:getMKCOfSelectedCharacter()
	return self.MotifKnowledgeContainers[self.SelectedCharacter]
end

function CharacterDataManager:getMKCfOCharacter(characterName)
	return self.MotifKnowledgeContainers[characterName]
end

function CharacterDataManager:getNonSelectedCharactersWithResearch(resKey, comparisonCharactersOnly)
	local known = {}
	local unknown = {}
	local compChars = {}
	local tcc = self:getComparisonCharacters()
	for i = 1, #tcc do
		compChars[tcc[i]] = 1
	end
	compChars[API.GetCharacterName()] = 1
	
	for charName, rpc in pairs(self.ResearchProgressContainers) do
		if (self.SelectedCharacter ~= charName) then
			if (comparisonCharactersOnly and (compChars[charName] ~= nil)) then
				local rpi = rpc:getByResearchKey(resKey)
				if (rpi.status ~= RPI.StatusType.UNKNOWN) then
					table.insert(known, charName)
				else
					table.insert(unknown, charName)
				end
			end
		end
	end
	return known, unknown
end

function GetLibResearchCharacterDataManager()
	return CharacterDataManager
end
